@@ -1,35 +1,31 @@
 #!/usr/bin/env python
-# -*- coding: utf-8 -*-
 # emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
 # vi: set ft=python sts=4 ts=4 sw=4 et:
-"""
-sMRIPrep: Structural MRI PREProcessing workflow
-===============================================
-"""
+"""sMRIPrep: Structural MRI PREProcessing workflow."""
 
 
 def main():
-    """Entrypoint"""
+    """Set an entrypoint."""
     opts = get_parser().parse_args()
     return build_opts(opts)
 
 
 def check_deps(workflow):
+    """Make sure all dependencies are installed."""
     from nipype.utils.filemanip import which
     return sorted(
         (node.interface.__class__.__name__, node.interface._cmd)
         for node in workflow._get_all_nodes()
-        if (hasattr(node.interface, '_cmd') and
-            which(node.interface._cmd.split()[0]) is None))
+        if (hasattr(node.interface, '_cmd')
+            and which(node.interface._cmd.split()[0]) is None))
 
 
 def get_parser():
-    """Build parser object"""
+    """Build parser object."""
     from pathlib import Path
     from argparse import ArgumentParser
     from argparse import RawTextHelpFormatter
-    from templateflow.api import templates
-    from .utils import ParseTemplates
+    from niworkflows.utils.spaces import Reference, SpatialReferences, OutputReferencesAction
     from ..__about__ import __version__
 
     parser = ArgumentParser(description='sMRIPrep: Structural MRI PREProcessing workflows',
@@ -55,6 +51,11 @@ def get_parser():
     g_bids.add_argument('--participant-label', '--participant_label', action='store', nargs='+',
                         help='a space delimited list of participant identifiers or a single '
                              'identifier (the sub- prefix can be removed)')
+    g_bids.add_argument(
+        '--bids-filter-file', action='store', type=Path, metavar='PATH',
+        help='a JSON file describing custom BIDS input filters using pybids '
+             '{<suffix>:{<entity>:<filter>,...},...} '
+             '(https://github.com/bids-standard/pybids/blob/master/bids/layout/config/bids.json)')
 
     g_perfm = parser.add_argument_group('Options to handle performance')
     g_perfm.add_argument('--nprocs', '--ncpus', '--nthreads', '--n_cpus', '-n-cpus',
@@ -76,25 +77,26 @@ def get_parser():
 
     g_conf = parser.add_argument_group('Workflow configuration')
     g_conf.add_argument(
-        '--output-spaces', nargs='+', action=ParseTemplates,
-        help='paths or keywords prescribing standard spaces to which normalize the input T1w image'
-             ' (valid keywords are: %s).' % ', '.join('"%s"' % s for s in templates()))
+        '--output-spaces', nargs='*', action=OutputReferencesAction, default=SpatialReferences(),
+        help='paths or keywords prescribing output spaces - '
+             'standard spaces will be extracted for spatial normalization.')
     g_conf.add_argument(
         '--longitudinal', action='store_true',
         help='treat dataset as longitudinal - may increase runtime')
-    g_conf.add_argument(
-        '--template', '--spatial-normalization-target', action='store', type=str,
-        choices=[tpl for tpl in templates() if not tpl.startswith('fs')],
-        help='DEPRECATED: spatial normalization target (one TemplateFlow Identifier)')
 
     #  ANTs options
     g_ants = parser.add_argument_group('Specific options for ANTs registrations')
-    g_ants.add_argument('--skull-strip-template', action='store', default='OASIS30ANTs',
-                        choices=['OASIS30ANTs', 'NKI', 'MNI152NLin2009cAsym'],
-                        help='select ANTs skull-stripping template (default: OASIS30ANTs))')
+    g_ants.add_argument(
+        '--skull-strip-template', default='OASIS30ANTs', type=Reference.from_string,
+        help='select a template for skull-stripping with antsBrainExtraction')
     g_ants.add_argument('--skull-strip-fixed-seed', action='store_true',
                         help='do not use a random seed for skull-stripping - will ensure '
                              'run-to-run replicability when used with --omp-nthreads 1')
+    g_ants.add_argument(
+        '--skull-strip-mode', action='store', choices=('auto', 'skip', 'force'),
+        default='auto', help='determiner for T1-weighted skull stripping (force ensures skull '
+                             'stripping, skip ignores skull stripping, and auto automatically '
+                             'ignores skull stripping if pre-stripped brains are detected).')
 
     # FreeSurfer options
     g_fs = parser.add_argument_group('Specific options for FreeSurfer preprocessing')
@@ -102,6 +104,10 @@ def get_parser():
         '--fs-license-file', metavar='PATH', type=Path,
         help='Path to FreeSurfer license key file. Get it (for free) by registering'
              ' at https://surfer.nmr.mgh.harvard.edu/registration.html')
+    g_fs.add_argument(
+        '--fs-subjects-dir', metavar='PATH', type=Path,
+        help='Path to existing FreeSurfer subjects directory to reuse. '
+             '(default: OUTPUT_DIR/freesurfer)')
 
     # Surface generation xor
     g_surfs = parser.add_argument_group('Surface preprocessing options')
@@ -109,24 +115,16 @@ def get_parser():
                          help='disable sub-millimeter (hires) reconstruction')
     g_surfs_xor = g_surfs.add_mutually_exclusive_group()
 
-    g_surfs_xor.add_argument(
-        '--fs-output-spaces', required=False, action='store', nargs='+',
-        choices=['fsnative', 'fsaverage', 'fsaverage6', 'fsaverage5'],
-        help="""DEPRECATED - configure Freesurfer's output spaces:
-  - fsnative: individual subject surface
-  - fsaverage*: FreeSurfer average meshes
-this argument can be single value or a space delimited list,
-for example: --fs-output-spaces fsnative fsaverage fsaverage5."""
-    )
-    g_surfs_xor.add_argument('--fs-no-reconall', '--no-freesurfer',
+    g_surfs_xor.add_argument('--fs-no-reconall',
                              action='store_false', dest='run_reconall',
-                             help='disable FreeSurfer surface preprocessing.'
-                             ' Note : `--no-freesurfer` is deprecated and will be removed in 1.2.'
-                             ' Use `--fs-no-reconall` instead.')
+                             help='disable FreeSurfer surface preprocessing.')
 
     g_other = parser.add_argument_group('Other options')
     g_other.add_argument('-w', '--work-dir', action='store', type=Path, default=Path('work'),
                          help='path where intermediate results should be stored')
+    g_other.add_argument(
+        '--fast-track', action='store_true', default=False,
+        help='fast-track the workflow by searching for existing derivatives.')
     g_other.add_argument(
         '--resource-monitor', action='store_true', default=False,
         help='enable Nipype\'s resource monitoring to keep track of memory and CPU usage')
@@ -155,7 +153,7 @@ for example: --fs-output-spaces fsnative fsaverage fsaverage5."""
 
 
 def build_opts(opts):
-    """Entry point"""
+    """Trigger a new process that builds the workflow graph, based on the input options."""
     import os
     from pathlib import Path
     import logging
@@ -164,6 +162,7 @@ def build_opts(opts):
     import warnings
     from multiprocessing import set_start_method, Process, Manager
     from nipype import logging as nlogging
+    from niworkflows.utils.misc import check_valid_fs_license
 
     set_start_method('forkserver')
 
@@ -176,19 +175,17 @@ def build_opts(opts):
 
     warnings.showwarning = _warn_redirect
 
-    # FreeSurfer license
-    default_license = str(Path(os.getenv('FREESURFER_HOME')) / 'license.txt')
     # Precedence: --fs-license-file, $FS_LICENSE, default_license
-    license_file = Path(opts.fs_license_file or os.getenv('FS_LICENSE', default_license))
-    if not license_file.exists():
+    license_file = opts.fs_license_file or os.getenv('FS_LICENSE')
+    if not check_valid_fs_license(lic=license_file):
         raise RuntimeError(
             'ERROR: a valid license file is required for FreeSurfer to run. '
             'sMRIPrep looked for an existing license file at several paths, in this '
             'order: 1) command line argument ``--fs-license-file``; 2) ``$FS_LICENSE`` '
             'environment variable; and 3) the ``$FREESURFER_HOME/license.txt`` path. '
             'Get it (for free) by registering at https://'
-            'surfer.nmr.mgh.harvard.edu/registration.html')
-    os.environ['FS_LICENSE'] = str(license_file)
+            'surfer.nmr.mgh.harvard.edu/registration.html'
+        )
 
     # Retrieve logging level
     log_level = int(max(25 - 5 * opts.verbose_count, logging.DEBUG))
@@ -214,7 +211,6 @@ def build_opts(opts):
         plugin_settings = retval['plugin_settings']
         bids_dir = retval['bids_dir']
         output_dir = retval['output_dir']
-        work_dir = retval['work_dir']
         subject_list = retval['subject_list']
         run_uuid = retval['run_uuid']
         retcode = retval['return_code']
@@ -249,7 +245,7 @@ def build_opts(opts):
         if opts.run_reconall:
             from templateflow import api
             from niworkflows.utils.misc import _copy_any
-            dseg_tsv = str(api.get('fsaverage', suffix='dseg', extensions=['.tsv']))
+            dseg_tsv = str(api.get('fsaverage', suffix='dseg', extension=['.tsv']))
             _copy_any(dseg_tsv,
                       str(Path(output_dir) / 'smriprep' / 'desc-aseg_dseg.tsv'))
             _copy_any(dseg_tsv,
@@ -261,7 +257,7 @@ def build_opts(opts):
 
         logger.log(25, 'Writing reports for participants: %s', ', '.join(subject_list))
         # Generate reports phase
-        errno += generate_reports(subject_list, output_dir, work_dir, run_uuid,
+        errno += generate_reports(subject_list, output_dir, run_uuid,
                                   packagename='smriprep')
         write_derivative_description(bids_dir, str(Path(output_dir) / 'smriprep'))
     sys.exit(int(errno > 0))
@@ -269,59 +265,28 @@ def build_opts(opts):
 
 def build_workflow(opts, retval):
     """
-    Create the Nipype Workflow that supports the whole execution
-    graph, given the inputs.
+    Create the Nipype Workflow that supports the whole execution graph, given the inputs.
+
     All the checks and the construction of the workflow are done
     inside this function that has pickleable inputs and output
     dictionary (``retval``) to allow isolation using a
     ``multiprocessing.Process`` that allows smriprep to enforce
     a hard-limited memory-scope.
+
     """
-    import sys
     from shutil import copyfile
     from os import cpu_count
-    from collections import OrderedDict
     import uuid
     from time import strftime
     from subprocess import check_call, CalledProcessError, TimeoutExpired
     from pkg_resources import resource_filename as pkgrf
 
+    import json
     from bids import BIDSLayout
     from nipype import logging, config as ncfg
+    from niworkflows.utils.bids import collect_participants
     from ..__about__ import __version__
     from ..workflows.base import init_smriprep_wf
-    from niworkflows.utils.bids import collect_participants
-
-    # Set the default template to 'MNI152NLin2009c'
-    output_spaces = opts.output_spaces or OrderedDict([('MNI152NLin2009cAsym', {})])
-
-    if opts.template:
-        print("""\
-The ``--template`` option has been deprecated in version 1.1.2. Your selected template \
-"%s" will be inserted at the front of the ``--output-spaces`` argument list. Please update \
-your scripts to use ``--output-spaces``.""" % opts.template, file=sys.stderr)
-        deprecated_tpl_arg = [(opts.template, {})]
-        # If output_spaces is not set, just replate the default - append otherwise
-        if opts.output_spaces is not None:
-            deprecated_tpl_arg += list(output_spaces.items())
-        output_spaces = OrderedDict(deprecated_tpl_arg)
-
-    if opts.fs_output_spaces:
-        print("""\
-The ``--fs-output-spaces`` option has been deprecated in version 1.1.2. Your selected output \
-surfaces "%s" will be appended to the ``--output-spaces`` argument list. Please update \
-your scripts to use ``--output-spaces``.""" % ', '.join(opts.fs_output_spaces), file=sys.stderr)
-        for fs_space in opts.fs_output_spaces:
-            if fs_space not in output_spaces:
-                output_spaces[fs_space] = {}
-
-    FS_SPACES = set(['fsnative', 'fsaverage', 'fsaverage6', 'fsaverage5'])
-    if opts.run_reconall and not list(FS_SPACES.intersection(output_spaces.keys())):
-        print("""\
-Although ``--fs-no-reconall`` was not set, no FreeSurfer output space (valid values are: \
-%s) was selected. Adding default "fsaverage5" to the \
-list of output spaces.""" % ', '.join(FS_SPACES), file=sys.stderr)
-        output_spaces['fsaverage5'] = {}
 
     logger = logging.getLogger('nipype.workflow')
 
@@ -330,6 +295,8 @@ list of output spaces.""" % ', '.join(FS_SPACES), file=sys.stderr)
       * BIDS dataset path: {bids_dir}.
       * Participant list: {subject_list}.
       * Run identifier: {uuid}.
+
+    {spaces}
     """.format
 
     # Set up some instrumental utilities
@@ -341,6 +308,8 @@ list of output spaces.""" % ', '.join(FS_SPACES), file=sys.stderr)
     subject_list = collect_participants(
         layout, participant_label=opts.participant_label)
 
+    bids_filters = json.loads(opts.bids_filter_file.read_text()) if opts.bids_filter_file else None
+
     # Load base plugin_settings from file if --use-plugin
     if opts.use_plugin is not None:
         from yaml import load as loadyml
@@ -427,34 +396,38 @@ list of output spaces.""" % ', '.join(FS_SPACES), file=sys.stderr)
         logger.log(25, 'Running --reports-only on participants %s', ', '.join(subject_list))
         if opts.run_uuid is not None:
             run_uuid = opts.run_uuid
-        retval['return_code'] = generate_reports(
-            subject_list, str(output_dir), str(work_dir), run_uuid,
-            config=pkgrf('smriprep', 'data/reports/config.json'))
+        retval['return_code'] = generate_reports(subject_list, str(output_dir), run_uuid,
+                                                 packagename="smriprep")
         return retval
 
-    # Build main workflow
     logger.log(25, INIT_MSG(
         version=__version__,
         bids_dir=bids_dir,
         subject_list=subject_list,
-        uuid=run_uuid)
+        uuid=run_uuid,
+        spaces=opts.output_spaces)
     )
 
+    # Build main workflow
     retval['workflow'] = init_smriprep_wf(
         debug=opts.sloppy,
+        fast_track=opts.fast_track,
         freesurfer=opts.run_reconall,
+        fs_subjects_dir=opts.fs_subjects_dir,
         hires=opts.hires,
         layout=layout,
         longitudinal=opts.longitudinal,
         low_mem=opts.low_mem,
         omp_nthreads=omp_nthreads,
         output_dir=str(output_dir),
-        output_spaces=output_spaces,
         run_uuid=run_uuid,
         skull_strip_fixed_seed=opts.skull_strip_fixed_seed,
-        skull_strip_template=opts.skull_strip_template,
+        skull_strip_mode=opts.skull_strip_mode,
+        skull_strip_template=opts.skull_strip_template[0],
+        spaces=opts.output_spaces,
         subject_list=subject_list,
         work_dir=str(work_dir),
+        bids_filters=bids_filters,
     )
     retval['return_code'] = 0
 
